# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import os
import uno, unohelper
from datetime import datetime, timedelta

from com.sun.star.ui.dialogs.TemplateDescription import FILEOPEN_SIMPLE
from com.sun.star.ui.dialogs.CommonFilePickerElementIds import PUSHBUTTON_OK
from com.sun.star.awt.MessageBoxType import MESSAGEBOX, INFOBOX, ERRORBOX, WARNINGBOX, QUERYBOX
from com.sun.star.awt.MessageBoxButtons import (BUTTONS_OK, BUTTONS_YES_NO, BUTTONS_YES_NO_CANCEL,
                                                DEFAULT_BUTTON_NO, DEFAULT_BUTTON_CANCEL)
from com.sun.star.sdbc.DataType import INTEGER
from com.sun.star.sdbc.ColumnValue import NULLABLE
from com.sun.star.sdb.application.DatabaseObject import FORM
from com.sun.star.uno import RuntimeException

#Les variables suivantes sont déclarées dans __init__.py :
# connexion = None
# databasedocument = None
# colonnesPos = []


class Office(object):
    '''Frequently used methods in office context'''
    class BarreDeProgression:
        '''Crée une barre de progresion'''
        def __init__(self, indicator, txt, max):
            self.indicator = indicator
            self.txt = txt
            self.max = max
        def setValue(self, value):
            self.indicator.setValue(value)
        def setText(self, txt):
            self.indicator.setText(txt)
        def setTextValue(self, txt, value):
            self.indicator.setText(txt)
            self.indicator.setValue(value)
        def start(self):
            self.indicator.start(self.txt, self.max)
        def end(self):
            self.indicator.end()
        def reset(self):
            self.indicator.reset()
        
    def __init__(self):
        self.ctx = uno.getComponentContext()
        self.smgr = self.ctx.ServiceManager
        self.frame = self.getDesktop().getCurrentFrame()
        self.extID = "org.openoffice.fbsc.lesablier.tooournoi"

    def createUnoService(self, service):
        return self.smgr.createInstance(service)

    def createUnoDialog(self,url):
        dp = self.createUnoService("com.sun.star.awt.DialogProvider")
        return dp.createDialog(url)

    def getDesktop(self):
        return self.smgr.createInstanceWithContext("com.sun.star.frame.Desktop",self.ctx)

    def getCurrentComponent(self):
        return self.getDesktop().getCurrentComponent()

    def getCurrentFrame(self):
        return self.getDesktop().getCurrentFrame()

    def getCurrentController(self):
        return self.frame.getController()
        
    def executeDispatch(self, command, args=None, frame=None):
        if not frame:
            frame = self.frame
        if not args:
            args=()
        dispatcher = self.createUnoService("com.sun.star.frame.DispatchHelper")
        dispatcher.executeDispatch(frame, command, "", 0, args)

    def createPropertyValues(self, **kwargs):
        '''Chaque paramètre nommé doit prendre la forme nom=valeur :
        createPropertyValues(Name=Value, Name=Value,...) -> (PropertyValue, PropertyValue,...)'''
        from com.sun.star.beans import PropertyValue
        return tuple(PropertyValue(k,0,kwargs[k],0) for k in kwargs)

    def hasUnoInterface(self, obj, interface):
        return interface in (t.typeName for t in obj.Types)

    def loadComponentFromURL(self, url, target="_default", searchflags=0, args=()):
        return self.getDesktop().loadComponentFromURL(url, target, searchflags, args)
    
    def convertFromUrl(self, url):
        return unohelper.fileUrlToSystemPath(url)

    def convertToUrl(self, url):
        try:
            nurl = unohelper.systemPathToFileUrl(url)
        except RuntimeException:
            nurl = unohelper.systemPathToFileUrl(unohelper.fileUrlToSystemPath(url))
        return nurl

    def getRegistryKeyContent(self, sKeyName, bforUpdate=False):
        '''Adapted from OpenOffice basic tools'''
        oConfigProvider = self.createUnoService("com.sun.star.configuration.ConfigurationProvider")
        aNodePath = self.createPropertyValues(nodepath=sKeyName)
        if bforUpdate:
            return oConfigProvider.createInstanceWithArguments("com.sun.star.configuration.ConfigurationUpdateAccess", aNodePath)
        else:
            return oConfigProvider.createInstanceWithArguments("com.sun.star.configuration.ConfigurationAccess", aNodePath)

    def substitutePath(self, url, bResubstitute=False):
        '''Adapted from OpenOffice basic tools.
        Traduit les variables de répertoire utilisées par OOo'''
        oPS = self.createUnoService("com.sun.star.util.PathSubstitution")
        if bResubstitute:
            return oPS.reSubstituteVariables(url)
        else:
            return oPS.substituteVariables(url, True)

    def filePicker(self, filters=(("Tous les fichiers (*.*)", "*.*"),), multi=False):
        """Instancie le service uno filepicker"""
        selFichier = self.createUnoService("com.sun.star.ui.dialogs.FilePicker")
        selFichier.initialize((FILEOPEN_SIMPLE,))
        for f in filters:
            selFichier.appendFilter(*f)
        selFichier.setMultiSelectionMode(multi)
        selFichier.setLabel(PUSHBUTTON_OK, "Choisir")
        if selFichier.execute():
            return selFichier.getSelectedFiles()
        return None

    def msgbox(self, message, titre="Message", boxtype='message', boutons=('ok',)):
        types = {'message': MESSAGEBOX, 'info': INFOBOX, 'error': ERRORBOX,
                 'warn': WARNINGBOX, 'query': QUERYBOX}
        btn = {'ok': BUTTONS_OK, 'yesno': BUTTONS_YES_NO, 'yesnocancel': BUTTONS_YES_NO_CANCEL,
               'defno': DEFAULT_BUTTON_NO,'defcancel': DEFAULT_BUTTON_CANCEL}
        tk = self.createUnoService("com.sun.star.awt.Toolkit")
        win = self.frame.ContainerWindow
        box = tk.createMessageBox(win, types[boxtype], sum(btn[x] for x in boutons), titre, message)
        return box.execute()

    def barreDeProgression(self, txt, max=100):
        indicator = self.frame.createStatusIndicator()
        return Tournoi.BarreDeProgression(indicator,txt,max)

    def xray(self, obj=None):
        '''Macro to call Basic XRay by Bernard Marcelly from Python.
        If no object is given, the current document's selection is used'''
        sURL = "vnd.sun.star.script:XRayTool._Main.Xray?language=Basic&location=application"
        if not obj:
            obj = self.getCurrentController().getSelection()
        oMSPF = self.createUnoService("com.sun.star.script.provider.MasterScriptProviderFactory")
        oMSP = oMSPF.createScriptProvider('')
        oScript = oMSP.getScript(sURL)
        oScript.invoke((obj,), (), ())

    def iterUno(self, obj):
        '''Permet l'itération sur certains objets UNO.
        Interfaces actuellement supportées : XEnumeration, XNameAccess et XIndexAccess.
        Exemple : iterUno(spreadsheet.Sheets) -> <generator object iterUno>'''
        if hasattr(obj,'getByIndex'):
            for n in range(obj.Count): 
                yield obj.getByIndex(n)
        elif hasattr(obj,'getByName'):
            for n in obj.ElementNames:
                yield obj.getByName(n)
        elif hasattr(obj,'createEnumeration'):
            enum = obj.createEnumeration() 
            while enum.hasMoreElements():
                yield enum.nextElement()
        else:
            raise TypeError("iterUno() ne fonctionne pas sur ce type d'objet.")

    def docSiOuvert(self,URL):
        '''Renvoie l'objet document correspondant à l'URL si ce document est en cours d'édition'''
        '''Utilisé pour mettre à jour les classeurs classent s'ils sont ouverts pour édition'''
        Desktop = self.getDesktop()
        for c in self.iterUno(Desktop.Components):
            try:
                if c.URL==URL: return c
            except AttributeError:
                pass
        return None

    def calcdate2datetime(self,ooodate):
        '''Traduit la date numérique d'un classeur Calc en un objet datetime python
        calcdate2datetime(numdate) -> datetime object'''
        basedate = self.getRegistryKeyContent("org.openoffice.Office.Calc/Calculate/Other/Date", False)
        return datetime(basedate.YY,basedate.MM,basedate.DD)+timedelta(ooodate)

    def TextToClipboard(self,cText):
        '''Thanks to DannyB, from http://www.oooforum.org/forum/viewtopic.phtml?t=6065
        Given a string of text, put that text into the clipboard.
        This works by...
           1. Invisibly create a new Writer document
           2. Insert text into that document.
           3. Select All
           4. Copy
           5. Close the invisible writer document.
        So what ends up in the clipboard can be pasted to an external
        program as Text.  But the clipboard contains additional information
        about the text, such as whatever default font and style it had in
        the temporary Writer document.'''
        args = self.createPropertyValues(Hidden=True)
        oDoc = self.loadComponentFromURL( "private:factory/swriter", "_blank", 0, args )
        # Get the text of the document.
        oText = oDoc.getText()
        # Get a cursor that can move over or to any part of the text.
        oCursor = oText.createTextCursor()
        # Insert text and paragraph breaks into the text, at the cursor position.
        oText.insertString( oCursor, cText, False )
        # frame = oDoc.CurrentController.Frame
        self.executeDispatch(".uno:SelectAll")
        self.executeDispatch(".uno:Copy")
        oDoc.close( True )


class Tournoi(Office):
    '''Fonctions de base pour la bonne marche de tOOournoi'''
    def __init__(self):
        super(Tournoi,self).__init__()
        # le document actif
        self.doc = self.getCurrentComponent()
        # récupère la connexion à la base
        # self.setTournoiConnection()
        # self.connexion = self.connexions['connexion']
        self.stmt = connexion.createStatement()
        self.execute = self.stmt.execute
        # infos générales
        self.starttext = "intitulé de la compétition"
        self.infos = {}
        self.infos['nom'] = self.getTournoiInfo('NomTournoi')
        self.infos['tablesaumerite'] = self.getTournoiInfo('TablesAuMerite','Boolean')
        self.infos['tablesparalleles'] = self.getTournoiInfo('TablesParalleles','Boolean')
        self.infos['suspredistr'] = self.getTournoiInfo('SuspRedistr','Boolean')
        self.infos['paires'] = self.getTournoiInfo('Paires','Boolean')
        self.infos['maxpartie'] = self.getTournoiInfo('PartiesMax', 'Int')
        self.infos['derpartie'] = self.getTournoiInfo('PartiesJouees', 'Int')
        self.infos['symbolefixe'] = self.getTournoiInfo('SymboleFixe')
        self.infos['tooournoiPath'] = self.getTOOournoiPath()
        self.infos['extPath'] = self.getExtPath()
        self.infos['extVer'] = self.getExtVer()

    def getTournoiInfo(self, column, type='String'):
        '''Renvoie les métadonnées du tournoi par le nom de la colonne dans la table T_Infos'''
        rs = self.executeQuery('SELECT "%s" FROM "T_Infos"' % column)
        rs.next()
        return getattr(rs, 'get%s' % type)(1)

    def genereRowSet(self, iType, table, filtre=''):
        '''Routine renvoyant un rowset.
        >>> genereRowSet(type commande, document source, filtre) -> objet RowSet'''
        oRS = self.createUnoService("com.sun.star.sdb.RowSet")
        oRS.CommandType = iType
        oRS.Command = table
        oRS.ActiveConnection = connexion
        oRS.Filter = filtre
        oRS.ApplyFilter = bool(filter)
        return oRS

    def executeQuery(self,sql):
        '''Exécute la requête SQL fournie en paramètre et retourne les données'''
        return self.stmt.executeQuery(sql)

    def nomDuJoueur(self, licence):
        '''Retourne le nom d'un joueur à partir de son numéro de licence'''
        req = 'select "Prénom" || \' \' || "Nom" from "T_Joueurs" where "Licence" = %s' % licence
        rowset = self.executeQuery(req)
        rowset.next()
        if rowset.Row:
            return rowset.getString(1)
        else:
            return ''

    def createTooournoiDialog(self, dialogue):
        '''Charge et retourne un dialogue créé en Basic'''
        return self.createUnoDialog("vnd.sun.star.script:tOOournoi.%s?location=application" % dialogue)

    def inputbox(self, texte='', valeur=''):
        dlg = self.createTooournoiDialog('inputbox')
        if texte:
            dlg.getControl('texte').setText(texte)
        cvaleur = dlg.getControl('valeur')
        cvaleur.setText(valeur)
        sel = cvaleur.Selection
        sel.Max = len(cvaleur.Text)
        cvaleur.Selection = sel
        if dlg.execute():
            return dlg.getControl('valeur').getText()
        else:
            return ''

    def clesOptions(self):
        '''Renvoie le registre de configuration de tOOournoi'''
        return self.getRegistryKeyContent("org.openoffice.tooournoi.Config", True)

    def getTOOournoiPath(self):
        '''Renvoie le répertoire de travail de tOOournoi'''
        cles = self.clesOptions()
        sPath = cles.Path
        if sPath[-1] != "/": sPath += "/"
        return self.substitutePath(sPath + cles.SubPath + "/")

    def getExtPath(self):
        '''Renvoie le répertoire d'installation de tOOournoi'''
        pip = self.ctx.getByName("/singletons/com.sun.star.deployment.PackageInformationProvider")
        extPath = pip.getPackageLocation(self.extID)
        if extPath[-1] != "/": extPath += "/"
        return extPath

    def getExtVer(self):
        '''Fonction renvoyant la version de tOOournoi.'''
        pip = self.ctx.getByName("/singletons/com.sun.star.deployment.PackageInformationProvider")
        extensions = pip.getExtensionList()
        for e in extensions:
            if self.extID in e:
                return e[1]
        return ''

    def resetTournoi(self,nbparties):
        '''Remet à zéro valeurs et données pour démarrer un nouveau tournoi'''
        # nettoie les fichiers de classement, qui seront obsolètes, et le sous-répertoire correspondant si vide
        dir = self.convertFromUrl('%(tooournoiPath)s%(nom)s' % self.infos)
        for f in ('Cumul','Partie','Inscrits','Stats'):
            try:
                os.remove('%s/%s_%s.ods' % (dir,self.infos['nom'],f))
            except OSError:  # le fichier n'existe pas
                pass
        try:
            if not os.listdir(dir): os.rmdir(dir)
        except OSError:
            pass
        # archive les données
        self.cree_archives()
        # efface les données de parties
        self.executeQuery( 'DELETE FROM "T_Parties"' )
        # efface les données d'information
        self.executeQuery( 'DELETE FROM "T_Infos"' )
        # réinitialise la table d'information
        self.executeQuery( 'INSERT INTO "T_Infos" VALUES (0,\'%s\',1,0,0,0,1,0,null,\'*\',0,null,null,null,0,null)' % self.starttext)
        # réinitialise la table des données de parties
        self.executeQuery( 'ALTER TABLE "T_Parties" ALTER COLUMN "ID_Partie" RESTART WITH 0' )
        # insère la ligne des tops dans la table T_Parties
        self.executeQuery( 'INSERT INTO "T_Parties" ("Licence", "Fixe", "DateInscription", '
                           '"Infos", "TableProvisoire", "TableEnCours") VALUES (0,null,null,0,0,0)' )
        # réinitialise la table Acteurs
        sSQL = ( 'DELETE FROM "T_Acteurs";\n'
                 'ALTER TABLE "T_Acteurs" ALTER COLUMN "ID_Acteur" RESTART WITH 0;')
        self.executeQuery( sSQL )
        ##actualise la source de données (commenté, semble superflu, à vérifier)
        #actualiserTables
        # réinitialise les tables et requêtes à une seule partie
        self.modifierNombreParties(nbparties)

    def estInscrit(self, licence):
        '''Retourne l'inscription d'un joueur à partir de sa licence'''
        rowset = self.executeQuery('SELECT "ID_Partie" FROM "T_Parties" WHERE "Licence" = %s' % licence)
        rowset.next()
        if not rowset.Row:
            rowset = self.executeQuery('SELECT "ID_Partie" FROM "T_Parties" WHERE "Licence2" = %s' % licence)
            rowset.next()
            if not rowset.Row:
                return False
        rowset.dispose()
        return True

    def cree_archives(self):
        '''Archives les donnéesen créant une copie des tables'''
        if self.infos['nom'] == self.starttext: return
        sSql = ('DROP TABLE "T_Parties_{0}" IF EXISTS;\n'
                'DROP TABLE "T_Infos_{0}" IF EXISTS;\n'
                'DROP TABLE "T_Acteurs_{0}" IF EXISTS;\n'
                'SELECT * INTO "T_Parties_{0}" FROM "T_Parties";\n'
                'SELECT * INTO "T_Infos_{0}" FROM "T_Infos";\n'
                'SELECT * INTO "T_Acteurs_{0}" FROM "T_Acteurs";\n'
                'ALTER TABLE "T_Parties_{0}" ADD PRIMARY KEY ("ID_Partie");\n'
                'ALTER TABLE "T_Infos_{0}" ADD PRIMARY KEY ("ID_Infos");\n'
                'ALTER TABLE "T_Acteurs_{0}" ADD PRIMARY KEY ("ID_Acteur");').format(self.infos['nom'])
        self.executeQuery( sSql )

    def modifierNombreParties(self, nb):
        '''Modifie le nombre de colonnes 'parties' dans la table T_Parties'''
        def txtStatuts(n):
            # la table du dernier 'from' est sans importance, si elle contient une ligne au moins
            where = ' > 0 AND "Cumul" > 0' if n > 1 else '> 0'
            txtSQL = ('select {0} as "Parties", '
                      '(select "T{0}" = 0 from "T_Parties" where "ID_Partie" = 0) as "Initiée", '
                      'case (select count("P{0}") from "T_Parties" where "ID_Partie" {1}) '
                      'when 0 then null '
                      'when (select count("ID_Partie") from "T_Parties" where "ID_Partie" {1}) then \'COMPLETS\' '
                      'else \'incomplets\' end as "Résultats" from "T_Infos"').format(n, where)
            return txtSQL

        table = connexion.Tables.getByName("T_Parties")
        ancParties = (table.Columns.Count-10)//3   # 10 = les dix premières colonnes, non modifiées
        seqCumul = ['coalesce("P1",0)']
        seqCalc_cumul = [', "P1", "R1"']
        seqStatuts = [txtStatuts(1)]

        for i in range(2,nb+1):
            seqCalc_cumul.append(', "P{0}", "R{0}"'.format(i))
            seqCumul.append('+ coalesce("P{0}",0)'.format(i))
            seqStatuts.append(txtStatuts(i))

        split1 = connexion.Queries.getByName("Calc_cumul").Command.split(" FROM")
        split2 = split1[0].split('"Cumul"')
        split2[-1] = ''.join(seqCalc_cumul)
        split1[0] = '"Cumul"'.join(split2)
        commande = " FROM".join(split1)
        connexion.Queries.getByName("Calc_cumul").Command = commande

        split1 = connexion.Queries.getByName("Calc_résultats").Command.split(" FROM")
        split2 = split1[0].split('"Cumul"')
        split2[-1] = ''.join(seqCalc_cumul)
        split1[0] = '"Cumul"'.join(split2)
        commande = " FROM".join(split1)
        connexion.Queries.getByName("Calc_résultats").Command = commande

        connexion.Queries.getByName("R_statuts").Command = "\nunion\n".join(seqStatuts)

        if 1 < nb < ancParties:
            for i in range(table.Columns.Count -1, 9+(nb*3), -1):
                table.Columns.dropByIndex(i)
        elif nb > ancParties:
            aCols = ("T","R","P")
            testDD = table.Columns.createDataDescriptor()
            testDD.Type = INTEGER
            testDD.IsNullable = NULLABLE
            testDD.Precision = 10
            testDD.DefaultValue = ""
            for i in range(ancParties+1, nb+1):
                for col in aCols:
                    testDD.Name = '%s%d' % (col,i)
                    table.Columns.appendByDescriptor(testDD)

    def ouvreForm(self, nom, foredition=False):
        '''Routine ouvrant un formulaire.'''
        if nom == "F_Inscriptions":
            nom += "_paires" if self.infos['paires'] else "_solo"
        return databasedocument.CurrentController.loadComponent(FORM, nom, foredition)

    def forms_reload(self):
        '''Met à jour les informations affichées dans le menu principal'''
        try:
            forms = databasedocument.FormDocuments.getByName("Menu_Principal").Component.DrawPage.Forms
        except AttributeError:  # le menu principal est édjà fermé
            return
        # modifie affichage résumé des inscrits
        bool = self.infos['paires']
        form = forms.getByName("Standard_Menu_Inscriptions")
        tc = form.TableControl
        tc.ColClub.Hidden = bool
        tc.ColInscrits.Hidden = bool
        tc.ColJoueur1.Hidden = not(bool)
        tc.ColJoueur2.Hidden = not(bool)
        #recharge tous les formulaires et sous-formumlaires
        for form in self.iterUno(forms):
            form.reload()

    def fixePositionEtTaille(self, source):
        '''Fixe les dimensions des divers formulaires et masque les barres d'outils.'''
        self.removeToolbars()
        coordict = {"F_Inscriptions_solo": (200,50,650,740,15), "F_Joueurs": (240,170,775,650,15),
                    "F_Acteurs": (300,70,700,700,15), "F_Permutation": (235,150,785,715,15),
                    "F_Saisie_Points": (380,100,475,835,15), "Menu_Options": (415,150,430,520,15),
                    "F_Inscriptions_paires": (200,50,900,730,15)}
        for k in coordict:
            if k in source.Title:
                coord = coordict[k]
                break
        else:
            coord = (190,150,880,520,15)
        ctrl = source.CurrentController
        ctrl.Frame.ContainerWindow.setPosSize(*coord)
        ctrl.ViewSettings.ZoomValue = 100
        ctrl.ViewSettings.ShowVertScrollBar = False
        self.doc.CurrentController.Frame.ContainerWindow.setEnable(True)

    def removeToolbars(self):
        '''Masque les barres d'outils.
        Adapté de l'extension Switchboard de Roberto Benitez
        (http://extensions.services.openoffice.org/project/SwitchBoard)'''
        Frame = self.doc.CurrentController.Frame
        Layout = Frame.LayoutManager
        for I in Layout.Elements:
            Layout.hideElement(I.ResourceURL)
        # et accessoirement les panneaux latéraux éventuels
        for pname in ("Navigator", "Sidebar", "DesignerDialog"):
            args = self.createPropertyValues(**{pname:False})
            self.executeDispatch(".uno:%s")

    def importViaCalc(self, url, Hidden=False, coldef=''):
        oTypeDetect = self.createUnoService("com.sun.star.document.TypeDetection")
        props = self.createPropertyValues(URL=url)
        sType = oTypeDetect.queryTypeByDescriptor(props, True)
        if not sType:
            return None
        elif "Text" in sType[0]:
            # charge les données dans Calc
            props = self.createPropertyValues(FilterOptions = "59/44,34/39,ANSI,1,%s" % coldef,
                                              FilterName = "Text - txt - csv (StarCalc)",
                                              Hidden=Hidden)
            return self.loadComponentFromURL(url, args=props)
        elif "calc" in sType[0] or "excel" in sType[0].lower():
            props = self.createPropertyValues(Hidden=Hidden)
            return self.loadComponentFromURL(url, args=props)
        else:
            return None

    def dataarray2csv(self, path, dataarray, sep=',', encoding="utf-8"):
        """
        Routine simplifiée d'export des résultats en csv
        @path: destination de sauvegarde
        @dataarray: plage de données à traiter
        """
        import codecs
        csv = []
        try:
            for row in dataarray:
                # calc ne connaît que le format 'double', d'où cette ligne :-(
                csv.append(sep.join(str(int(r)) if isinstance(r,float) else r for r in row))
            with codecs.open(path, 'wb', encoding=encoding) as f:
                f.write('\n'.join(csv))
            return True
        except Exception as err:
            self.msgbox("Erreur: {0}".format(err))
            return False
